//---------------------------------------------------------------------------
// SMS Check (c) Omar Cornut (Bock) 2000-2014
// Sega 8-bit ROM tool, showing checksum and various informations
//---------------------------------------------------------------------------
// Note:
// - This is considering outdated code. Prefer using Maxim's Header Reader:
//   http://www.smspower.org/maxim/utilities/headerreader.html
//   I am maintaining this tool for personal usage.
//---------------------------------------------------------------------------
// Compilation:
// # gcc *.c
// # gcc *.c -DUNIX     // under UNIX system
//---------------------------------------------------------------------------
// History:
// 2000/10/11 - v0.10
// - Initial release.
// 2000/10/24 - v0.20
// 2006/04/15 - v0.21 
// - Added heuristic mapper detection code (from MEKA).
// - Added full range sum printing (to compare with marking on EPROM).
// - Renamed to "SMS Check" not to confuse with Maxim's tool.
// 2014/11/29 - v0.22
//  - Rebuilding for modern Windows/etc.
//  - Fixed typos.
//  - Disabled mapper detection code.
//---------------------------------------------------------------------------

#define PROG_NAME       "SMS Check"
#define AUTHOR          "Omar Cornut (Bock)"
#define VERSION         "0.22"
#define WEB_PAGE        "http://www.smspower.org"

#include <fcntl.h>
#include "win_dirent.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

//---------------------------------------------------------------------------

#include "liblist.h"
#include "liblist.c"

typedef unsigned char   u8;
typedef unsigned short  u16;

static int	    Match(const char *src, const char *wildcards);

//---------------------------------------------------------------------------

typedef struct
{
    char  Sega_ID[8];
    u16   Unknown_W1; // Protocol ?
    u16   Checksum;
    u16   PartNo;     // Product Number
    u8    Version;
    u8    System_and_Size;
} t_smsheader;

// Systems:
//  0x30  Master System Japan
//  0x40  Master System Export
//  0x50  Game Gear Japan
//  0x60  Game Gear International
//  0x70  Game Gear Export

// Sizes: (used for checksum)
//  0x0A  8k
//  0x0B  16k
//  0x0C  32k
//  0x0D  48k
//  0x0E  64k
//  0x0F  128k
//  0x00  256k

//---------------------------------------------------------------------------

void    smschecker_process_file(const char* filename)
{
    printf ("\n[%s]\n", filename);
	FILE* f = fopen(filename, "rb");
    if (!f)
	{
        printf("Error opening file!\n");
        return;
    }

	fseek(f, 0, SEEK_END);
	const long filesize = ftell(f);
	fseek(f, 0, SEEK_SET);

    printf("Size ...... %li\n", filesize);
    if (filesize < 0x8000)
    {
        printf("File is too short to contain a header!\n");
        fclose(f);
        return;
    }

    u8* filebuf = (u8*)malloc(filesize);
	{
		int r = fread(filebuf, 1, filesize, f);
		if (r != filesize)
		{
			printf("Unable to read file. Tried to read %d bytes, got %d.\n", filesize, r);
			free(filebuf);
			fclose(f);
			return;
		}
	}

	t_smsheader   smsheader;
    memcpy(&smsheader, filebuf + 0x8000 - sizeof (smsheader), sizeof (smsheader));

    // HEADER -------------------------------------------------------------------
    {
        printf ("Header ....");
        for (int i = 0; i < sizeof (smsheader); i++)
            printf (" %02X", ((u8 *)(&smsheader))[i]);
        printf ("\n");
    }

    // SYSTEM -------------------------------------------------------------------
    {
        u8 System = (smsheader.System_and_Size & 0xF0) >> 4;
        printf ("System .... ");
        switch (System)
        {
        case 3:  printf ("SMS Japan");        break;
        case 4:  printf ("SMS Export");       break;
        case 5:  printf ("GG Japan");         break;
        case 6:  printf ("GG International"); break;
        case 7:  printf ("GG Export");        break;
        default: printf ("Unknown");          break;
        }
        printf (" (%i)\n", System);
    }

    // SIZE ---------------------------------------------------------------------
	int checksum_len = 0;
    {
        u8 Size = (smsheader.System_and_Size & 0x0F);
        printf ("Size ...... ");
        switch (Size)
        {
        case 0x0A: checksum_len = 0x02000;  printf ("8k");      break;
        case 0x0B: checksum_len = 0x04000;  printf ("16k");     break;
        case 0x0C: checksum_len = 0x08000;  printf ("32k");     break;
        case 0x0D: checksum_len = 0x0C000;  printf ("48k");     break;
        case 0x0E: checksum_len = 0x10000;  printf ("64k");     break;
        case 0x0F: checksum_len = 0x20000;  printf ("128k");    break;
        case 0x00: checksum_len = 0x40000;  printf ("256k+");   break;
        default:   checksum_len = filesize; printf ("Unknown"); break;
        }
        printf (" \n");
    }

    // CHECKSUM -----------------------------------------------------------------
    {
        u16 Checksum;
        unsigned int Fullsum;
        int ret;

        printf ("CheckSum .. ");
        if (filesize < checksum_len)
        {
            printf ("File too short!\n");
        }
        else
        {
            // Compute checksum
            Checksum = 0;
            for (int i = 0; i < checksum_len; i++)
                Checksum += filebuf [i];
            // Remove header from checksum
            if (checksum_len >= 0x8000)
                for (int i = 0x8000 - sizeof (smsheader); i < 0x8000; i++)
                    Checksum -= filebuf [i];
            // Print out result
            printf ("%04X, ", Checksum);
            if (smsheader.Checksum == Checksum)
                printf ("Ok\n");
            else
                printf ("Bad. Should be %04X\n", smsheader.Checksum);
        }

        printf ("FullSum ... ");
        Fullsum = 0;
        for (int i = 0; i < filesize; i++)
          Fullsum += filebuf [i];
        printf ("%06X\n", Fullsum);
    }

	/*
    // MAPPER -------------------------------------------------------------------
    {
        int c8000 = 0;
        int cFFFF = 0;

        // Search for mapper accessing code
        for (int i = 0; i < 0x8000; i++)
            if (filebuf [i] == 0x32) // Z80 Opcode: LD (xxxx), A
            {
                if (filebuf[i + 1] == 0x00 && filebuf[i + 2] == 0x80)
                { i += 2; c8000++; }
                else
                    if (filebuf[i + 1] == 0xFF && filebuf[i + 2] == 0xFF)
                    { i += 2; cFFFF++; }
            }

            // Print out result
            printf ("Mapper .... %02d/%02d -> %s", c8000, cFFFF, (c8000 > cFFFF) ? "CodeMasters" : "Normal");
    }
	*/

    free(filebuf);
    fclose(f);
}

void    smschecker_process_files(char *FileMask)
{
    char           *Path, *Files, *p;
    DIR            *Dir;
    struct dirent  *DirEnt;
    t_list         *FileList;

    FileMask = strdup (FileMask);

#ifndef UNIX
    {
        char *s = FileMask;
        char c;
        while ((c = *s) != 0)
        {
            if (c == '\\')
                *s = '/';
            s++;
        }
    }
#endif

    if ((p = strrchr(FileMask, '/')))
    {
        Path = FileMask;
        Files = p + 1;
        *p = '\0';
    }
    else
    {
        Path = ".";
        Files = FileMask;
    }
#ifndef UNIX
    strupr(Files);
#endif

    if (!(Dir = opendir(Path)))
    {
        free (FileMask);
        return;
    }

    FileList = 0;
    while ((DirEnt = readdir(Dir)))
    {
#ifdef UNIX
        char *FileName = DirEnt->d_name;
#else
        char *FileName = strdup(DirEnt->d_name);
        strupr(FileName);
#endif
        if (Match(FileName, Files) > 0)
        {
            char* FullFileName = (char*)malloc(sizeof(char) * (strlen(Path) + 1 + strlen(DirEnt->d_name) + 1));
            sprintf(FullFileName, "%s/%s", Path, DirEnt->d_name);
            list_add (&FileList, FullFileName);
        }
#ifndef UNIX
        free(FileName);
#endif
    }
    if (FileList)
    {
        t_list *files;
        list_sort (&FileList, (int (*)(void *, void *))strcmp);
        for (files = FileList; files != NULL; files = files->next)
        {
            smschecker_process_file((const char*)files->elem);
        }
        list_free (&FileList);
    }
    else
    {
        printf ("Sorry, but no files matches with \"%s\".\n", FileMask);
    }

    closedir (Dir);
    free (FileMask);
}

int     main (int ArgC, char **ArgV)
{
    int   i;

    printf (PROG_NAME " " VERSION " - by " AUTHOR ", " __DATE__ "\n" WEB_PAGE "\n--\n");
    if (ArgC < 2)
    {
        printf ("Syntax: %s [romfile] [romfile] ..\n", ArgV[0]);
        printf ("(wildcards allowed)\n");
        return 0;
    }
    for (i = 1; i < ArgC; i++)
        smschecker_process_files(ArgV[i]);
    return 0;
}

//---------------------------------------------------------------------------

static int	    Match (const char *src, const char *wildcards)
{
    int	nbr;
    int	val;
    int	i;

    nbr = 0;
    for (val = 0; ; val++)
    {
        if (wildcards[val] == '*')
        {
            i = 0;
            do
            {
                nbr = nbr + Match(src + val + i, wildcards + val + 1);
                i++;
            }
            while (src[val + i - 1] != '\0');
            return (nbr);
        }
        else
        {
            if (wildcards[val] != src[val])
                return (0);
            if (wildcards[val] == '\0')
                return (1);
        }
    }
}

//---------------------------------------------------------------------------
